# Object-oriented healpy wrapper with support for multi-resolutions maps

import healpy as hp

import HealpixMap as hmap

import matplotlib.pyplot as plt

import operator

from copy import copy,deepcopy

from astropy.io import fits
from astropy.table import Table

import numpy as np
from numpy import array, log2, sqrt

import collections

from .healpix_base import HealpixBase

class HealpixMap(HealpixBase):
    """
    Object-oriented healpy wrapper with support for multi-resolutions maps 
    (known as multi-order coverage map, or MOC).

    You can instantiate a map by providing either:
    
    * Size (through ``order`` or ``nside``), and a ``scheme`` ('RING' or 'NESTED').
      This will initialize an empty map.
    * An array (in ``data``) and an a scheme ('RING' or 'NESTED'). This will
      initialize the contents of the single-resolution map.
    * A structured array or ``recarray``. This will initialize a MOC map. The 
      scheme 'NUNIQ' is implicit in this case.

    .. warning::
        The initialization input is not validated by default. Consider calling 
        `is_mesh_valid()` after initialization, otherwise results might be
        unexpected.


    Regardless of the underlaying grid, you can operate on maps using 
    ``*``, ``/``, ``+``, ``-``, ``**``, ``==``, ``abs`` and ``floor``. For binary 
    operations the result always corresponds to the finest grid, so there
    is no loss of information. If any of the operands is a MOC, the result is a
    MOC with an appropiate updated grid.. If both operands have the same NSIDE, 
    the scheme of the result corresponds to the left operand. If you want to 
    preserve the grid for a specific operand, use ``*=``, `/=`, etc. 

    .. warning::
        Information might degrade if you use in-place operators (e.g. ``*=``, `/=`)

    The maps are array-like, that is, the can be casted into a regular numpy 
    array (as used by healpy), are iterable (over the pixel values) and can be
    used with built-in function such as ``sum`` and ``max``. For MOC maps, if
    you want to keep the `UNIQ` pixel numbers in a ``recarray``, you have to use
    ``data()``.

    You can also access the value of pixels using regular numpy indexing 
    with ``[]``. For MOC maps, no specific pixel ordering is guaranted. For a 
    given pixel number ``ipix`` in the current grid, you can get the 
    corresponding UNIQ pixel number using ``m.pix2uniq(ipix)``.

    Args:
        data (array): Value for an initialize map. A MOC map is a 2-fields 
            recarray whose first field is called 'UNIQ'
        order (int): Order of HEALPix map.
        nside (int): Alternatively, you can specify the NSIDE parameter.
        scheme (str): Healpix scheme. Either 'RING', 'NESTED' or 'NUNIQ'
        density (bool): Whether the value of each pixel should be treated as
            counts in a histogram (``False``) or as the value of a [density]
            function evaluated at the center of the pixel (``True``). This affect 
            operations involving the splitting of a pixel. 
        dtype (array): Numpy data type. Will be ignored if data is provided.
    """

    def __init__(self,
                 data = None,
                 order = None,
                 nside = None,
                 scheme = 'ring',
                 density = False,
                 dtype = None):
        """
        Initialize an empty (use either nside or order) or initialized 
        """

        if data is not None:
            # Initializes map contents

            # Standarize to array
            data = array(data)

            # Single or multi resolution?
            if (np.size(data.dtype.names) == 2 and
                data.dtype.names[0] == "UNIQ"):
                # Interpret as a multi-order coverage (MOC) map

                data = data.view(np.recarray)

                super().__init__(uniq = data.field(0))

                self._data = data.field(1)
                
            else:
                # Single resolution map

                order = hp.nside2order(hp.npix2nside(data.size))

                super().__init__(order = order,
                                 scheme = scheme)

                self._data = data

        else:
            # Empty map
            
            super().__init__(nside = nside,
                             order = order,
                             scheme = scheme)

            self._data = np.zeros(self.npix, dtype=dtype)

        # Other properties
        self._density = density
            
    @classmethod
    def read_map(cls, filename, field = None, uniq_field = 0, hdu = 1,
                 density = False):
        """
        Read a HEALPix map from a FITS file.

        Args:
            filename (Path): Path to file
            field (int): Column where the map contents are. Default: 0 for 
                single-resolution maps, 1 for MOC maps.
            uniq_field (int): Column where the UNIQ pixel numbers are. 
                For MOC maps only. 
            hdu (int): The header number to look at. Starts at 0.
            density (bool): Whether this is a histogram-like or a density-like map.
        
        Return:
            HealpixMap
        """

        with fits.open(filename) as hdul:

            hdu = hdul[hdu]

            scheme = hdu.header["ORDERING"]

            if scheme == 'NUNIQ':
                # Is MOC

                if field is None:
                    field = 1

                uniq = hdu.data.field(uniq_field)
                contents = hdu.data.field(field)

                data = np.rec.fromarrays([uniq, contents],
                                 names = ['UNIQ', 'CONTENTS'])

                m = cls(data, density = density)

            else:
                # Sigle resolution

                if field is None:
                    field = 0

                m = cls(hdu.data.field(field), scheme=scheme)

        return m

    def write_map(self,
                  filename,
                  extra_maps = None,
                  column_names = None,
                  extra_header = None,
                  overwrite = False):
        """
        Write map to disc.

        Args:
            filename (Path): Path to output file
            extra_maps (HealpixMap or array): Save more maps in the same file
                as extra columns. Must be conformable.
            column_names (str or array): Name of colums. Must have the same 
                length as the number for maps. Defaults to 'CONTENTSn', where 
                ``n`` is the map number (ommited for a single map). For MOC maps,
                the pixel information is always stored in the first column, called
                'UNIQ'.
            extra_header (iterable): Iterable of (keyword, value, [comment]) tuples
            overwrite (bool): If True, overwrite the output file if it exists. 
                Raises an OSError if False and the output file exists. 
        """
        
        # Standarize data for astropy's Table
        if self.is_moc:

            data = [self._uniq, self._data]

        else:

            data = [self._data]

        # Add extra maps
        if extra_maps is not None:
            
            if isinstance(extra_maps, HealpixMap):
                extra_maps = (extra_maps,)

            for map in extra_maps:

                if not self.conformable(map):
                    raise ValueError("All extra maps must be conformable")

                data.append(map._data)

        # Column names
        nmaps = len(data) - self.is_moc
        
        if column_names is not None:

            if isinstance(column_names, str):
                column_names = [column_names]
            
            if len(column_names) != nmaps:
                raise ValueError("Colum names must match the number of maps.")
            
        else:

            if nmaps > 1:

                column_names = ["CONTENTS{}".format(i) for i in range(nmaps)]
                
            else:

                column_names = ["CONTENTS"]

        if self.is_moc:

            column_names.insert(0, 'UNIQ')

        # Header
        header = [('PIXTYPE', 'HEALPIX',
                       'HEALPIX pixelisation'),
                  ('ORDERING', self.scheme,
                       'Pixel ordering scheme: RING, NESTED, or NUNIQ'),
                  ('NSIDE', self.nside,
                       'Resolution parameter of HEALPIX'),
                  ('INDXSCHM', 'EXPLICIT' if self.is_moc else 'IMPLICIT',
                       'Indexing: IMPLICIT or EXPLICIT')]

        if extra_header is not None:
            header.extend(extra_header)
        
        # Prepare table and write
        table = Table(data, names = column_names)

        hdu = fits.table_to_hdu(table)

        hdu.header.extend(header)

        hdulist = fits.HDUList([fits.PrimaryHDU(), hdu])

        hdulist.writeto(filename, overwrite = overwrite)
        
    
    @classmethod
    def adaptive_moc_map(cls, pix_fun, combine_fun, max_order, density = False):
        """
        Discretize a function into an adaptive multi-order map.

        This method evaluates ``pix_fun`` for all pixels in a map of order
        ``max_order`` and then, guided by ``combine_fun``, recursively groups 
        together sets of 4 pixels into a single parent pixel of a lower order.

        Args:
            pix_fun (function): Function to be discretize. It shall take two 
                arguments: ``nside`` (``int``) and ``pix`` (``int`` or ``array``),
                and return the value of the function being discretized a for the
                input pixels (all of which correspond to the same ``nside``). The
                output should have the same size as the input.
            combine_fun (function): Function that determines 4 pixels contained
                by a single pixel at lower order should be combined. It shall
                take three arguments: ``nside`` (``int``), ``pix`` 
                (``array`` of 4 elements) and the corresponding ``values`` 
                (``array``).
                All pixels correspond to the same ``nside``. The function returns
                ``False`` if these four pixels should remain split, and `True`
                if they should be combined into a lower order.
            max_order (int): The function will be initially sampled at this order
            density (bool): If ``True``, the value for pixels at lower order will
                obtaine by evaluating ``pix_fun`` again, otherwise it will be the
                sum of the values of the child pixels (i.e. the map is
                histogram-like and the operation is a rebinning)

        Return:
            HealpixMap
        """

        # Cache multiple values for each order
        order_list = array(range(max_order, -1, -1))
        nside_list = 2**order_list
        uniq_shift_list = 4 * nside_list * nside_list
        npix_ratio_list = 4 ** (max_order - order_list)
        next_npix_ratio_list = 4**array(range(1,max_order+2))

        # We've reached the base pixels, there's no next order.
        # This value is needed as a stopping condition later on
        next_npix_ratio_list[-1] = next_npix_ratio_list[-2]

        # Initialize empty map
        map_uniq = array([], dtype = int)
        map_data = array([])

        # We'll use the rangeset representation and will convert to NUNIQ later
        # on. We only need to keep in memory 4 pixels at a time for each order,
        # and nbuffer keep track of how many we currently have
        start = np.zeros((max_order+1, 4), dtype=int)
        stop = np.zeros((max_order+1, 4), dtype=int)
        value = np.zeros((max_order+1, 4))
        nbuffer = np.zeros((max_order+1), dtype=int)

        # Loop through all pixels at max_order, 4 at a time
        for pix in range(0, hmap.order2npix(max_order), 4):

            # Initial evaluation at finer pixelation
            next_pix = pix + 4

            pix_list = array(range(pix, next_pix))

            start[0] = pix_list
            stop[0] = pix_list + 1
            value[0] = pix_fun(nside_list[0], pix_list)
            nbuffer[0] = 4

            # Check the status at each order and either combine or add to the map
            for     step,(    nside,      uniq_shift,      npix_ratio,      next_npix_ratio) in \
                enumerate(zip(nside_list, uniq_shift_list, npix_ratio_list, next_npix_ratio_list)):

                if next_pix % next_npix_ratio == 0:
                    # We got through all of the children pixels already
                    
                    current_nbuffer = nbuffer[step]

                    pix_values = value[step][:current_nbuffer]

                    if step == max_order or current_nbuffer != 4 or not combine_fun(nside, start[step], pix_values):
                        # Add pixels to map
                        # Either
                        # 1. We've reached the base pixels
                        # 2. The parent pixel needs to be splitted because
                        #    there is structure at higher orders
                        # 3. They don't meet the requirements of combine_fun

                        uniq = (start[step][:current_nbuffer] /
                                npix_ratio + uniq_shift).astype(int) 
                        
                        map_uniq = np.append(map_uniq, uniq)
                        map_data = np.append(map_data, pix_values)

                    else:
                        # Combine into lower order pixel

                        next_step = step + 1
                        next_nbuffer = nbuffer[next_step]

                        next_start = start[step][0]
                        next_stop = stop[step][-1]

                        if density:
                            # Best approximated by evaluating the function at the centers
                            next_value = pix_fun(nside_list[next_step], int(next_start/next_npix_ratio))
                        else:
                            # If histogram-like, this is just rebinning
                            next_value = sum(pix_values)

                        start[next_step][next_nbuffer] = next_start
                        stop[next_step][next_nbuffer] = next_stop
                        value[next_step][next_nbuffer] = next_value
                        nbuffer[next_step] += 1

                    # Clear buffer at this order
                    nbuffer[step] = 0

        map_data = np.rec.fromarrays([map_uniq, map_data],
                                     names = ['UNIQ', 'CONTENTS'])
                    
        return cls(map_data, density = density)

    @classmethod
    def adaptive_moc_mesh(cls, split_fun, max_order, density = False,
                      dtype = None):
        """
        Return a zero-initialized MOC map, with an adaptive resolution
        determined by an arbitrary function.

        Args:
            split_fun (function): This method should return ``True`` if a pixel 
            should be split into pixel of a higher order, and ``False`` otherwise. 
            It takes two integers, ``start`` (inclusive) and ``stop`` (exclusive), 
            which correspond to a single pixel in nested rangeset format for a 
            map of order ``max_order``.
            max_order (int): Maximum HEALPix order to consider
            density (bool): Will be pass to HealpixMap initialization.
            dtype (dtype): Data type

        Return:
            HealpixMap
        """
        
        base = HealpixBase.adaptive_moc_mesh(split_fun, max_order)

        data = np.rec.fromarrays([base._uniq, np.zeros(base.npix, dtype = dtype)],
                                 names = ["UNIQ", "CONTENTS"])

        return cls(data, density = density)
        
    @classmethod
    def moc_from_pixels(cls, pixels, order, nest = False, density = False,
                        dtype = None):
        """
        Return a zero-initialize MOC map where a list of pixels are kept at a 
        given order, and every other pixel is appropiately downsampled.
        
        Also see the more generic ``adaptive_moc_map()`` and ``adaptive_moc_mesh()``.

        Args:
            pixels (array): Pixels that must be kept at the finest pixelation
            order (int): Maximum healpix order (that is, the order for the pixel
                list)
            nest (bool): Whether the pixels are a 'NESTED' or 'RING' scheme
            density (bool): Wheather the map is density-like or histogram-like
            dtype: Daty type
        """

        base = HealpixBase.moc_from_pixels(pixels, order, nest = nest)

        data = np.rec.fromarrays([base._uniq, np.zeros(base.npix, dtype = dtype)],
                                 names = ["UNIQ", "CONTENTS"])

        return cls(data, density = density)

    @classmethod
    def moc_histogram(cls, nside, samples, max_value, nest = False, weights = None):
        """
        Generate an adaptive MOC map by histogramming samples. 

        If the number of samples is greater than the number of pixels in a map
        of the input ``nside``, consider generating a single-resolution
        map and then use `to_moc()`.

        Also see the more generic ``adaptive_moc_map()`` and ``adaptive_moc_mesh()``.

        Args: 
            nside (int): Healpix NSIDE of the samples and maximum NSIDE of the 
                output map
            samples (int array): List of pixels representing the samples. e.g.
                the output of `healpy.ang2pix()`.
            max_value: maximum number of samples (or sum of weights) per pixel.
                Note that due to limitations of the input ``nside``, the output
                could contain pixels with a value largen than this
            nest (bool): Whether the samples are in NESTED or RING scheme
            weights (array): Optionally weight the samples. Both must have the
                same size.
        
        Return:
            HealpixMap
        """

        # Standarize samples
        if not nest:
            samples = array([hp.ring2nest(nside, pix) for pix in samples])


        if weights:
            samples = np.rec.fromarrays([samples, weights],
                                        names = ['pix', 'wgt'])

        else:
            samples = np.rec.fromarrays([samples],
                                 names = ['pix'])

        samples.sort(order = 'pix')

        # Get empty mesh by reusing adaptive_moc_mesh
        def value_fun(start, stop):

            start_pos, stop_pos = np.searchsorted(samples.pix, [start, stop])

            if weights:
                value = sum(samples.wgt[start_pos:stop_pos])
            else:
                value = stop_pos - start_pos

            return value
                
        order = hp.nside2order(nside)

        if weights:
            dtype = array(weights).dtype
        else:
            dtype = int
        
        moc_map = cls.adaptive_moc_mesh(lambda i,f: value_fun(i,f) > max_value,
                                        order,
                                        density = False,
                                        dtype = dtype)

        # Fill
        rangesets = moc_map.pix_rangesets(order)

        for pix,(start,stop) in enumerate(rangesets):
            
            moc_map[pix] = value_fun(start, stop)

        return moc_map

        
    def to_moc(self, max_value):
        """
        Convert a single-resolution map into a MOC based on the maximum value
        a given pixel the latter should have. 

        ... note:: 
            
            The maximum order of the MOC map is the same as the order of the 
            single-resolution map, so the output map could contain pixels with 
            a value greater than this.

        Also see the more generic ``adaptive_moc_map()`` and ``adaptive_moc_mesh()``.

        Args:
            max_value: Maximum value per pixel of the MOC. Whether the map is
                histogram-like or density-like is taken into account.

        Return:
            HealpixMap
        """

        if self.is_moc:
            return TypeError("Map is already a MOC")
        
        # Get empty mesh by reusing adaptive_moc_mesh
        def value_fun(start, stop):
            value = sum(self._data[start:stop])

            if self._density:
                value /= stop-start

            return value

        if self._density:
            def split_fun(start, stop):
                return max(self._data[start:stop]) > max_value
        else:
            def split_fun(start,stop):
                return value_fun(start,stop) > max_value
        
        moc_map = self.adaptive_moc_mesh(split_fun,
                                         self.order,
                                         density = self._density,
                                         dtype = self.dtype)

        # Fill
        rangesets = moc_map.pix_rangesets(self.order)

        for pix,(start,stop) in enumerate(rangesets):
            
            moc_map[pix] = value_fun(start, stop)

        return moc_map

    def density(self, density = None):
        """
        Set or get the density parameter.

        Args:
            density (bool or None): Whether the value of each pixel should be treated as
                counts in a histogram (``False``) or as the value of a [density]
                function evaluated at the center of the pixel (``True``). This affect 
                operations involving the splitting of a pixel. ``None`` will leave 
                this paramter unchanged.

        Return:
            bool: The current density

        """
        if density:
            self._density = density

        return self._density

    
    def data(self, contents_label = 'CONTENTS', uniq_label = 'UNIQ'):
        """
        Get the raw data in the form of an array.

        Args:
            contents_label (str): Name of the contents column (2nd), 
                for multi-resolution maps.
            uniq_label (str): Name of the UNIQ pixel numbers column (1st),
                for multi-resolution maps.

        Return:
            array or recarray: For single and multi-resolution maps, respectively.
        """

        if self.is_moc:

            return np.rec.fromarrays([self._uniq, self._data],
                                     names = [uniq_label, contents_label])
            
        else:

            return self._data
        
    @property
    def dtype(self):
        return self._data.dtype
        
    def __eq__(self, other):

        return (super().__eq__(self, other) and
                array_equal(self._data, other._data) and
                self._density == other._density)
                
    def __getitem__(self, key):

        return self._data[key]
        
    def __setitem__(self, key, value):

        self._data[key] = value
        
    def __imul__(self, other):

        return self._ioperation(other, operator.imul)

    def __mul__(self, other):

        return self._operation(other, operator.mul)

    def __rmul__(self, other):
            
        return self._operation(other, operator.mul)

    def __itruediv__(self, other):

        return self._ioperation(other, operator.itruediv)

    def __truediv__(self, other):

        return self._operation(other, operator.truediv)

    def __rtruediv__(self, other):
            
        return self._operation(other, operator.truediv)

    def __ifloordiv__(self, other):

        return self._ioperation(other, operator.ifloordiv)

    def __floordiv__(self, other):

        return self._operation(other, operator.floordiv)

    def __rfloordiv__(self, other):
            
        return self._operation(other, operator.floordiv)

    def __iadd__(self, other):

        return self._ioperation(other, operator.iadd)

    def __add__(self, other):

        return self._operation(other, operator.add)

    def __radd__(self, other):
            
        return self._operation(other, operator.add)

    def __isub__(self, other):

        return self._ioperation(other, operator.isub)

    def __sub__(self, other):

        return self._operation(other, operator.sub)

    def __rsub__(self, other):
            
        return self._operation(other, operator.sub)

    def __ipow__(self, other):

        return self._ioperation(other, operator.ipow)

    def __pow__(self, other):

        return self._operation(other, operator.pow)

    def __rpow__(self, other):
            
        return self._operation(other, operator.pow)

    def __neg__(self):

        new = deepcopy(self)

        new._data *= -1

        return new

    def __abs__(self):

        new = deepcopy(self)

        new._data = np.abs(new._data)
        
        return new
    
    def __array__(self):

        return self._data
        
    def rasterize(self, nside, scheme):
        """
        Convert to map of a given NSIDE and scheme

        Args:
            nside (int): HEALPix NSIDE
            scheme (str): RING or NESTED

        Return:
            HealpixMap
        """

        scheme = scheme.upper()
        
        if scheme not in ['RING', 'NESTED']:
            raise ValueError("Scheme must be RING or NESTED")

        if self.nside == nside and self.scheme == scheme:
            # Same grid, nothing to do
            
            return deepcopy(self)

        else:
            # All other case can be handled with the identity operation
            
            raster = HealpixMap(nside = nside,
                                scheme = scheme,
                                density = self._density)

            raster._ioperation(self, lambda a,b: b)

            return raster
        
    def _operation(self, map1, operation):

        # If second is not a map (e.g. scalar or array), we'll keep the grid
        # If any map is MOC, result will be MOC. Grid is updated.
        # If both are single-resolution, result will have the finest grid
        # If both are single resolution and same order, scheme will correspond
        # to first operand
        # If maps are conformable, the output grid remains unchanged

        map0 = self
        
        if np.isscalar(map1):
            # Operation by a scalar

            map0 = deepcopy(map0)
            return map0._ioperation(map1, operation)

        else:

            # Operation by another map or something that can be turned into a map

            # Cast to HealpixMap if needed
            if not isinstance(map1, HealpixMap):
                map1 = HealpixMap(map1)

            # Optimize for different cases
            if (map0.is_moc or map1.is_moc) and not map0.conformable(map1):
                
                # Multi-resolution map
                # This will change the underlaying NUNIQ grid

                # Convert pixel numbers to an equivalent sorted list of nested
                # rangeset for highest posible order
                max_order = max(map0.order, map1.order)

                rs0 = map0.pix_rangesets(max_order)
                rs1 = map1.pix_rangesets(max_order)

                sort0 = np.argsort(rs0.start)
                sort1 = np.argsort(rs1.start)

                # Initialize new data with highest possible number of pixels,
                # some will be discarded at the end but this is fasten than append
                new_uniq = np.zeros(map0.npix + map1.npix, dtype = int)
                new_data = np.zeros(map0.npix + map1.npix, dtype = map0._data.dtype)

                pos0 = 0
                pos1 = 0

                pix_new = 0

                start = 0 # Rangeset start of new pixel

                while pos0 < map0.npix and pos1 < map1.npix:

                    # Get input values
                    pix0   = sort0[pos0]
                    range0 = rs0[pix0]
                    value0 = map0[pix0] 

                    pix1   = sort1[pos1]
                    range1 = rs1[pix1]
                    value1 = map1[pix1]
                    
                    stop = min(range0.stop, range1.stop) # Rangeset stop of new pixel

                    # Handle density maps
                    npix_new = stop - start

                    if not map0._density:
                        npix_ratio0 = npix_new / (range0.stop-range0.start)
                        value0 = value0 * npix_ratio0

                    if not map1._density:
                        npix_ratio1 = npix_new / (range1.stop-range1.start)
                        value1 = value1 * npix_ratio1

                    # Operation
                    value = operation(value0, value1)

                    new_uniq[pix_new] = hmap.range2uniq(max_order,
                                                              (start, stop))
                    new_data[pix_new] = value

                    # Advance to next pixel
                    pix_new += 1

                    if stop == range0.stop:
                        pos0 += 1

                    if stop == range1.stop:
                        pos1 += 1

                    start = stop

                # Update density parameter
                # If any of the maps is histogram-like, the result is also a
                # weighted histogram-like map
                new_density = map0._density and map1._density

                # Create new map
                new_data = np.rec.fromarrays([new_uniq[:pix_new], new_data[:pix_new]],
                                              names = ["UNIQ", "CONTENTS"])

                return HealpixMap(new_data, density = new_density)
                
            else:

                # Single-resolution or conformable, can reuse in-place operation
                
                if map1.order > map0.order:
                    map0,map1 = map1,map0

                map0 = deepcopy(map0)
                    
                return map0._ioperation(map1, operation)

    
    def _ioperation(self, map1, operation):

        map0 = self
        
        if np.isscalar(map1):
            # Operation by a scalar

            map0._data = operation(map0._data, map1)

        else:

            # Operation by another map or something that can be turned into a map

            # Cast to HealpixMap if needed
            if not isinstance(map1, HealpixMap):
                map1 = HealpixMap(map1)

            # Optimize procedure for various situations
            if map0.conformable(map1):
                # Same underlaying grid, so easy operation

                map0._data = operation(map0._data, map1._data)

            elif ((map0.is_moc or map1.is_moc) or
                  ((map0.is_ring or map1.is_ring) and map0.order != map1.order)):

                # There is no clear optimization in this case
                # Will work in the rangeset representation

                max_order = max(map0.order, map1.order)

                rs0   = map0.pix_rangesets(max_order)
                sort0 = np.argsort(rs0.start)
                pos0  = 0

                rs1   = map1.pix_rangesets(max_order)
                sort1 = np.argsort(rs1.start)
                pos1  = 0

                while pos0 < map0.npix and pos1 < map1.npix:

                    pix0 = sort0[pos0]
                    range0 = rs0[pix0]
                    len0 = range0.stop - range0.start

                    pix1 = sort1[pos1]
                    range1 = rs1[pix1]
                    len1 = range1.stop - range1.start

                    if len0 > len1:
                        # Downgrade pix1 by getting summing over child pixels
                        # (or weighted average, for a density map)

                        value0 = map0[pix0]
                        value1 = 0

                        while True:
                            # Will break when we catch up with map0

                            if map1._density:
                                # Will take the weighted average
                                value1 += len1 * map1[pix1]
                            else:
                                # Simple sum
                                value1 += map1[pix1]

                            if range0.stop == range1.stop:
                                break

                            pos1   += 1
                            pix1   = sort1[pos1]
                            range1 = rs1[pix1]
                            len1   = range1.stop - range1.start

                        if map1._density:
                            value1 /= len0

                        map0[pix0] = operation(value0, value1)

                    else:

                        # Upgrade pix1 by dividing it up
                        # (or simply getting the value for density)

                        while True:
                            # Will break when we catch up with map1

                            value1 = map1[pix1]

                            if not map1._density:
                                value1 /= len1 // len0

                            value0 = map0[pix0]

                            map0[pix0] = operation(value0, value1)

                            if range0.stop == range1.stop:
                                break

                            pos0   += 1
                            pix0   = sort0[pos0]
                            range0 = rs0[pix0]
                            len0   = range0.stop - range0.start

                    pos0 += 1
                    pos1 += 1
                
                
            elif map0.order == map1.order:
                # Same order, different scheme, single-resolution

                nside = map0.nside

                if map0.scheme == 'NESTED':
                    # map1 is RING, map0 is NESTED

                    for pix in range(map0.npix):

                        map0[pix] = operation(map0[pix],
                                              map1[hp.nest2ring(nside, pix)])

                else:
                    # map1 is NESTED, map0 is RING

                    for pix in range(map0.npix):

                        map0[pix] = operation(map0[pix],
                                              map1[hp.ring2nest(nside, pix)])

            else:

                # Only possibility left is that both are NESTED with different
                # order, which is easy to handle

                if map1.order > map0.order:
                    # Downgrade map1 by summing or getting the weighted average

                    npix_ratio = int(4**(map1.order - map0.order))

                    for pix in range(map0.npix):

                        value1 = sum(map1._data[pix*npix_ratio:(pix+1)*npix_ratio])

                        if map1._density:
                            value1 /= npix_ratio

                        map0._data[pix] = operation(map0._data[pix], value1)

                else:

                    # Upgrade map1 by splitting the pixel
                    # (or just use the value if density)

                    npix_ratio = int(4**(map0.order - map1.order))

                    for pix in range(map1.npix):

                        value1 = map1._data[pix]

                        if not map1._density:
                            value1 /= npix_ratio

                        pix0_start = pix*npix_ratio
                        pix0_stop = pix0_start + npix_ratio

                        map0._data[pix0_start: pix0_stop] = \
                            operation(map0._data[pix0_start: pix0_stop],
                                      value1)

            # Update density parameter
            # If any of the maps is histogram-like, the result is also a
            # weighted histogram-like map
            map0._density = map0._density and map1._density

        return map0

    def plot(self,
             ax = None,
             proj = 'moll',
             rot=0,
             coord='C',
             flip='astro',
             xsize=800,
             ysize=None,
             lonra=[-180,180],
             latra=[-90,90],
             half_sky=False,
             reso=1.5,
             **kwargs):
        """
        Plot map. This is a wrapper for matplotlib.pyplot.imshow

        Args:
            ax (matplotlib.axes.Axes): Axes on where to plot the map. If ``None``,
                it will create a new figure.
            proj (str): Projections: 'moll' (molltweide), 'cart' (carthographic), 
                'orth' (orthographics) or 'gnom' (gnomonic)
            rot (float or sequence): Describe the rotation to apply. In the 
                form (lon, lat, psi) (unit: degrees) : the point at longitude 
                lon and latitude lat will be at the center. An additional 
                rotation of angle psi around this direction is applied. If a 
                scalar, the rotation is performed around zenith
            coord (str): Either one of ‘G’ (Galactic), ‘E’ (Equatorial) or 
                ‘C’ (Celestial) to describe the 
                coordinate system of the map, or a sequence of 2 of these to 
                rotate the map from the first to the second coordinate system.
            flip (str): Defines the convention of projection : ‘astro’ 
                (east towards left, west towards right) or ‘geo’ (east towards 
                right, west towards left)
            xsize (int): The horizontal size of the image.
            ysize (int): The verital size of the image. For carthographic and 
                gnomonic projections only.
            lonra (array): Range in longitude (degrees). For carthographic only. 
            latra (array): Range in latitude (degrees). For carthographic only.
            half_sky (bool): Plot only one side of the sphere. For orthographic 
                only
            reso (float): Resolution (in arcmin). For gnominic projection only.
            **kwargs: Passed to matplotlib.pyplot.imshow

        Return:
           AxesImage, healpix.projector.SphericalProj: The first return value
               corresponds to the output ``imgshow``. The second is the healpy's
               projector used. This is particularly useful to add extra elements
               to the plots, e.g.::

                   plot, proj = m.plot(ax, 'moll')
                   x,y = proj.ang2xy(np.deg2rad(90), np.deg2rad(45))
                   ax.text(x, y, "(zenith = 90 deg, azimuth = 45 deg)")
        """

        # Create axes if needed
        default_plot = False
        if ax is None:
            plt.figure(figsize=(8.5, 5.4))        
            ax = plt.gca()
            ax.axis('off')
            default_plot = True
            
        # Setup appropiate projection
        proj = proj.lower()
        
        if proj == 'moll':            
            projector = hp.projector.MollweideProj(rot=rot,
                                              coord=coord,
                                              flipconv=flip,
                                              xsize=xsize)
            
        elif proj == 'cart':
            projector = hp.projector.CartesianProj(rot=rot,
                                              coord=coord,
                                              flipconv=flip,
                                              xsize=xsize,
                                              ysize=ysize,
                                              lonra=lonra,
                                              latra=latra)
        elif proj == 'orth':
            projector = hp.projector.OrthographicProj(rot=rot,
                                                 coord=coord,
                                                 flipconv=flip,
                                                 xsize=xsize,
                                                 half_sky=half_sky)
        elif proj == 'gnom':
            projector = hp.projector.GnomonicProj(rot=rot,
                                             coord=coord,
                                             flipconv=flip,
                                             xsize=xsize,
                                             ysize=ysize,
                                             reso=reso)
            
        else:
            raise ValueError("Wrong porojection")
        
        # Get values
        class rasterizer:
            """
            This is a wrapper around [], that will divide the value of the pixel
            if the map is MOC and histogram-like. This will give the same result 
            as calling rasterize() first and then plotting the equivalent 
            single-resolution map
            """
            
            def __init__(self, map):
                self.map = map
            
            def __getitem__(self, pix):

                if self.map.is_moc and not self.map._density:

                    # Histogram-like MOC, will divide the pixel
                    pix_order = hmap.uniq2order(self.map._uniq[pix]) 

                    npix_ratio = 4 ** (self.map.order - pix_order)

                    return  self.map[pix] / npix_ratio
                    
                else:

                    # Single resolution or density MOC, nothing to do
                    
                    return self.map[pix]
                
        img = projector.projmap(rasterizer(self), self.vec2pix,
                                rot=rot, coord=coord)

        # Plot
        plot =  ax.imshow(img,
                          extent = projector.get_extent(),
                          origin="lower",
                          **kwargs)

        if default_plot:
            plt.gcf().colorbar(plot, orientation="horizontal")
        
        return plot, projector

    def get_interp_val(self, theta, phi):
        """
        Return the bi-linear interpolation value of a map using 4 nearest neighbours.

        For MOC maps, this is equivalent to raterizing the map first to the 
        highest order.

        Args:
            theta (float or array): Zenith angle (rad)
            phi (float or array): Azimuth angle (rad)

        Return:
            scalar or array
        """
        
        if self.is_moc:

            pixels,weights = self.get_interp_weights(theta, phi)

            values = self[pixels]
            
            if not self.density:
                # Split the pixels up to corresponding maximum order
                
                order = hmap.uniq2order(self._uniq[pixels])

                npix_ratio = 4 ** (self.order - order)

                values /= npix_ratio

            return sum(values * weights)
            
        else:

            return hp.get_interp_val(self._data, theta, phi, nest = self.is_nested)


        



# Pixelization related function for single resolutions maps

import healpy as hp

import numpy as np
from numpy import array

def order2npix(order):
    """
    Get the number of pixel for a map of a given order

    Args:
        order (int or array)

    Return:
        int or array
    """

    return 12*4**array(order, dtype=int)

# healpy functions not related to the grid , copied here just for convenience
vec2ang = hp.vec2ang
"""
Same as healpy.pixelfunc.vec2ang. Included here for convinience. 
"""

ang2vec = hp.ang2vec
"""
Same as healpy.pixelfunc.ang2vec. Included here for convinience. 
"""




# Useful plotting routines

import numpy as np

def plot_grid(m, ax, proj, step = 10, **kwargs):
    """
    Plot the pixel boundaries of a Healpix grid
    
    Args:
        m (HealpixBase): Map defining the grid
        ax (matplotlib.axes.Axes): Axes on where to plot
        proj (healpy.projector.SphericalProj): Projector to converto 
            spherical coordinates to plot's axes coodinates
        step (int): How many points per pixel side
        **kwargs: Passed to matplotlib.pyplot.plot()
    """
    
    # Keep track of limits
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()
    
    # Every line should have the same color
    if 'c' in kwargs:
        color = kwargs.pop('c')
    else:
        color = kwargs.pop('color', 'white')
    
    # Plot every pixel
    for pix in range(m.npix):

        # Get boundaries
        vec = m.boundaries(pix, step = step)
        
        # Close loop
        vec = np.append(vec, vec[:,0].reshape(-1,1), axis = 1)
 
        # Project
        x,y = proj.vec2xy(vec)

        # Remove discontinuities
        dx = np.abs(np.diff(x))
        dy = np.abs(np.diff(y))
        jumps = np.logical_or(dx > 2*np.mean(dx), 
                              dy > 2*np.mean(dy))
        jumps = np.append(jumps, False)

        if any(jumps):
            
            ijumps = np.nonzero(jumps)[0]+1
            
            x = np.insert(x, ijumps, np.nan)
            y = np.insert(y, ijumps, np.nan)
            
        # Plot
        ax.plot(x,y, color = color, **kwargs)

    # Set axes back to the original limits
    ax.set_xlim(*xlim)
    ax.set_ylim(*ylim)




import healpy as hp

import HealpixMap as hmap

import numpy as np
from numpy import array, log2, sqrt

class HealpixBase:
    """
    Basic operations related to HEALPix pixelization, for which the map
    contents information is not needed. This class is conceptually very similar 
    the the Healpix_Base class of Healpix_cxx.

    Single resolution maps are fully defined by specifying their order 
    (or NSIDE) and ordering scheme ("RING" or "NESTED"). 

    Multi-resolution maps follow an explicit "NUNIQ" scheme, with each pixel 
    identfied by a _uniq_ number. No specific is needed nor guaranteed.

    .. warning::
        The initialization input is not validated by default. Consider calling 
        `is_mesh_valid()` after initialization, otherwise results might be
        unexpected.


    Args:
        uniq (array): Explicit numbering of each pixel in an "NUNIQ" scheme.
        order (int): Order of HEALPix map.
        nside (int): Alternatively, you can specify the NSIDE parameter.
        scheme (str): Healpix scheme. Either 'RING', 'NESTED' or 'NUNIQ'
    """

    def __init__(self,
                 uniq = None,
                 order = None,
                 nside = None,
                 scheme = 'ring'):

        if uniq is not None:
            # MOC map

            self._uniq = uniq

            # Scheme and order are implicit
            self._scheme = "NUNIQ"
            
            self._order = hmap.uniq2order(max(uniq))
            
        else:
            # Single resolution map

            self._uniq = None
            
            # User specified nside instead of order
            if nside is not None:
                
                if order is not None:
                    raise ValueError("Specify either 'order' or 'nside'")
                
                order = hp.nside2order(nside)

            if order is None:
                raise ValueError("Specifcy nside or order")
                
            self._order = order

            if scheme is None:
                    raise ValueError("Specify scheme")

            self._scheme = scheme.upper()

            if self._scheme not in ['RING','NESTED', 'NUNIQ']:
                raise ValueError("Scheme can only be 'ring', 'nested' or 'NUNIQ'")


    def __eq__(self, other):

        if self.is_moc:
            return array_equal(self._uniq, other._uniq)
        else:
            return (self._order == other._order and 
                    self._scheme == other._scheme)
            
    def uniq(self, pix):
        """
        Get the UNIQ value of a pixel (as used in [])

        Args:
            pix (int or array): Pixel number

        Return:
            int
        """

        if self.is_moc:

            return self._uniq[pix]

        else:

            if not self.is_nested:
                pix = hp.ring2nest(self.nside, pix)
                
            return hmap.nest2uniq(self.nside, pix)
            
    @classmethod
    def adaptive_moc_mesh(cls, split_fun, max_order):
        """
        Return a MOC mesh with an adaptive resolution
        determined by an arbitrary function.

        Args:
            split_fun (function): This method should return ``True`` if a pixel 
            should be split into pixel of a higher order, and ``False`` otherwise. 
            It takes two integers, ``start`` (inclusive) and ``stop`` (exclusive), 
            which correspond to a single pixel in nested rangeset format for a 
            map of order ``max_order``.
            max_order (int): Maximum HEALPix order to consider

        Return:
            HealpixBase
        """

        map_uniq = array([], dtype = int)

        order_list = array(range(max_order+1))
        nside = 2**order_list
        npix_ratio = 4 ** (max_order - order_list)
        uniq_shift = 4 * nside * nside
        npix_ratio = 4 ** (max_order - order_list)

        start_buffer = np.zeros((max_order+1, 4), dtype=int)
        cursor = -np.ones((max_order+1), dtype=int)

        for base_pix in range(12):

            order = 0
            start_buffer[order][0] = base_pix * npix_ratio[order]
            cursor[order] = 0

            while order >= 0:

                if cursor[order] < 0:
                    order -= 1
                    continue

                start = start_buffer[order][cursor[order]]
                stop  = start + npix_ratio[order]

                cursor[order] -= 1

                if order < max_order and split_fun(start, stop):
                    # Split

                    order += 1

                    split_shift = array(range(4))*npix_ratio[order]
                    start_buffer[order] = start + split_shift

                    cursor[order] = 3

                else:
                    # Add to map

                    uniq = (start / npix_ratio[order] + uniq_shift[order]).astype(int) 

                    map_uniq = np.append(map_uniq, uniq)

        return cls(map_uniq)

    @classmethod
    def moc_from_pixels(cls, pixels, order, nest = False):
        """
        Return a MOC mesh where a list of pixels are kept at a 
        given order, and every other pixel is appropiately downsampled.
        
        Also see the more generic ``adaptive_moc()`` and ``adaptive_moc_mesh()``.

        Args:
            pixels (array): Pixels that must be kept at the finest pixelation
            order (int): Maximum healpix order (that is, the order for the pixel
                list)
            nest (bool): Whether the pixels are a 'NESTED' or 'RING' scheme
        """

        nside = hp.order2nside(order)

        # Always work in nested
        if not nest:
            pixels = array([hp.ring2nest(nside, pix) for pix in pixels])
            
        # Auxiliary function so we can reuse adaptive_moc_mesh()
        pixels.sort()
        
        def pix_list_range_intersect(start, stop):

            start_index,stop_index = np.searchsorted(pixels, [start,stop])

            return start_index < stop_index

        # Get the grid        
        return cls.adaptive_moc_mesh(pix_list_range_intersect, 
                                             order)

    def conformable(self, other):
        """
        For single-resolution maps, return ``True`` if both maps have the same
        order and scheme.

        For MOC maps, return `True` if both maps have the same list of UNIQ 
        pixels (including the ordering)
        """

        if self.scheme != other.scheme:
            return False
        
        if self.is_moc:

            return np.array_equal(self._uniq, other._uniq)

        else:

            return self.order == other.order

    @property
    def npix(self):
        """
        Get number of pixels.

        For multi-resolutions maps, this corresponds to the number of utilized 
        UNIQ pixels.

        Return:
            int
        """

        if self.is_moc:   
            return len(self._uniq)
        else:
            return int(12 * 4**self._order)

    @property
    def order(self):
        """
        Get map order

        Return:
            int
        """
        
        return self._order

    @property
    def nside(self):
        """
        Get map NSIDE

        Return:
            int
        """

        return int(2**self.order)
        
    @property
    def scheme(self):
        """
        Return HEALPix scheme
        
        Return:
            str: Either 'NESTED', 'RING' or 'NUNIQ'
        """

        return self._scheme

    @property
    def is_nested(self):
        """
        Return true if scheme is NESTED or NUNIQ
        
        Return
            bool
        """

        return self._scheme == "NESTED" or self._scheme == 'NUNIQ'
        
    @property
    def is_ring(self):
        """
        Return true if scheme is RING
        
        Return
            bool
        """

        return self._scheme == "RING"

    @property
    def is_moc(self):
        """
        Return true if this is a Multi-Dimensional Coverage (MOC) map 
        (multi-resolution)

        Return:
            bool
        """

        return self._uniq is not None

    def pix_rangesets(self, order):
        """
        Get the equivalent range of `child pixels` in nested scheme for a map 
        of a higher order

        Args:
            order (int): Order of output range sets

        Return:
            recarray: With columns named 'start' (inclusive) and 
                'stop' (exclusive) 
        """

        if self.scheme == 'NESTED':

            start,stop = hmap.nest2range(self.order, range(self.npix), order)

        elif self.scheme == 'RING':

            start,stop = hmap.nest2range(self.order,
                                         hp.ring2nest(self.nside, range(self.npix)),
                                         order)

        else:
            #self.scheme == 'NUNIQ'

            start,stop = hmap.uniq2range(self._uniq, order)

        return np.rec.fromarrays([start,stop], names = ['start', 'stop'])

    def pixarea(self, pix = 0):
        """
        Return area of pixel in steradians

        Args:
            pix (int or array): Pixel number. Only relevant for MOC maps
        
        Return:
            float or array
        """

        if self.is_moc:
            order = hmap.uniq2order(self._uniq[pix])
        else:
            order = self.order

        return 1.047197551196597746 / 4**order
        
    def pix2ang(self, pix):
        """
        Return the coordinates of the center of a pixel
        
        Args:
            pix (int or array)

        Return:
            (float or array, float or array)
        """

        if self.is_moc:

            pix,order = hmap.uniq2nest(self.pix2uniq(pix))

            nside = hp.order2nside(order)

            return hp.pix2ang(nside, pix, nest = True)
            
        else:
            return hp.pix2ang(self.nside, pix, nest = self.is_nested)
            
    def pix2vec(self, pix):
        """
        Return a vector corresponding to the center of a pixel
        
        Args:
            pix (int or array)

        Return:
            array: Size (3,N)
        """

        if self.is_moc:

            pix,order = hmap.uniq2nest(self.pix2uniq(pix))

            nside = hp.order2nside(order)

            return hp.pix2vec(nside, pix, nest = True)
            
        else:
            return hp.pix2vec(self.nside, pix, nest = self.is_nested)
            
    def ang2pix(self, theta, phi):
        """
        Get the pixel (as used in []) that contains a given coordinate

        Args:
            theta (float or array): Zenith angle
            phi (float or arrray): Azimuth angle
      
        Return:
            int or array
        """

        pixels = hp.ang2pix(self.nside, theta, phi, nest=self.is_nested)
        
        if self.is_moc:
            pixels = self.nest2pix(pixels)
            
        return pixels
            
    def vec2pix(self, x, y, z):
        """
        Get the pixel (as used in []) that contains a given coordinate

        Args:
            theta (float or array): Zenith angle
            phi (float or arrray): Azimuth angle
      
        Return:
            int or array
        """

        pixels = hp.vec2pix(self.nside, x, y, z, nest=self.is_nested)

        if self.is_moc:
            pixels = self.nest2pix(pixels)
            
        return pixels
        
    def pix2uniq(self, ipix):
        """
        Get the UNIQ representation of a given pixel index.

        Args:
            ipix (int): Pixel number in the current scheme (as used for [])
        """

        if self.is_moc:

            return self._uniq[ipix]

        else:

            if self.scheme == 'RING':
                pix = hp.ring2nest(self.nside, pix)

            return hmap.nest2uniq(self.nside, pix)

    def nest2pix(self, pix):
        """
        Get the corresponding pixel in the current grid for a pixel in NESTED
        scheme. For MOC map, return the pixel that contains it. 

        Args:
            pix (int or array): Pixel number in NESTED scheme. Must correspond 
                to a map of the same order as the current.

        Return:
            int or array
        """

        if self.is_moc:

            # Work with rangesets for maximum order,
            # then find pixel number for this order,
            # and then find the rangesets that contain these pixels

            rangesets = self.pix_rangesets(self.order)
            
            rs_argsort = np.argsort(rangesets.stop)
            
            ipix = np.searchsorted(rangesets.stop, pix,
                                   side = 'right',
                                   sorter = rs_argsort)

            opix = rs_argsort[ipix]

            opix[pix == -1] = -1 # Follow healpy convention for null pix

            return opix
            
        else:
            
            if self.is_nested:
                return pix
            else:
                return hp.nest2ring(self.nside, pix)
            
    
    def get_interp_weights(self, theta, phi):
        """
        Return the 4 closest pixels on the two rings above and below the 
        location and corresponding weights. Weights are provided for bilinear 
        interpolation along latitude and longitude

        Args:
            theta (float or array): Zenith angle (rad)
            phi (float or array): Azimuth angle (rad)
 
        Return:
            tuple: (pixels, weights), each with of (4,) if the input is scalar,
                if (4,N) where N is size of
                theta and phi. For MOC maps, these pixel numbers might repeate.
        """

        pixels,weights = hp.get_interp_weights(self.nside, theta, phi,
                                               nest = self.is_nested)

        if self.is_moc:
            pixels = self.nest2pix(pixels)

        return (pixels, weights)
            
    def get_all_neighbours(self, theta, phi = None):
        """
        Return the 8 nearest pixels. For MOC maps, these might repeat, as this
        is equivalent to raterizing the maps to the highest order, getting the 
        neighbohrs, and then finding the pixels tha contain them.

        Args:
            theta (float or int or array): Zenith angle (rad). If phi is 
                ``None``, these are assummed to be pixels numbers. For MOC maps,
                these are assumed to be pixel numbers in NESTED scheme for the
                equivalent single-resolution map of the highest order.
            phi (float or array or None): Azimuth angle (rad)

        Return:
            array: pixel number of the SW, W, NW, N, NE, E, SE and S neighbours,
                shape is (8,) if input is scalar, otherwise shape is (8, N) if 
                input is of length N. If a neighbor does not exist (it can be 
                the case for W, N, E and S) the corresponding pixel number will 
                be -1.
        """
        
        neighbors = hp.get_all_neighbours(self.nside, theta, phi,
                                          nest = self.is_nested)

        print(neighbors)
        
        if self.is_moc:
            neighbors = self.nest2pix(neighbors)

        return neighbors
 
    def is_mesh_valid(self):
        """
        Return ``True`` is the map pixelization is valid. For
        single resolution this simply checks that the size is a valid NSIDE value.
        For MOC maps, it checks that every point in the sphere is covered by
        one and only one pixel.
        
        Return:
            True
        """

        if self.is_moc:

            # Work in rangesets, and check that there is no gap in between them
            rs = self.pix_rangesets(self.order)

            rs.sort(order = 'start')

            return (rs.start[0] == 0 and
                    rs.stop[-1] == hp.nside2npix(self.nside) and
                    np.array_equal(rs.start[1:], rs.stop[:-1]))
            
        else:
                
            return hp.isnpixok(self.npix)

    def _pix_query_fun(self, fun):
        """
        Return a wrapper for healpy's pix querying routines

        Args:
            fun (function): Healpy's query_* functions
        
        Return:
            function: With apprpiate grid, passes rest of arguments to fun
        """

        def wrapper(*args, **kwargs):

            if self.is_moc:

                # We'll do it order by order

                pix_per_order = [[] for _ in range(self.order+1)]
                nest_pix_per_order = [[] for _ in range(self.order+1)]

                for pix in range(self.npix):

                    nest_pix,order = hmap.uniq2nest(self.uniq(pix))

                    pix_per_order[order].append(pix)
                    nest_pix_per_order[order].append(nest_pix)

                query_pix = np.zeros(0, dtype = int)

                for order in range(self.order+1):

                    pixels = fun(hp.order2nside(order), nest = True,
                                 *args, **kwargs)

                    query_bool = np.isin(nest_pix_per_order[order], pixels)

                    order_pix = array(pix_per_order[order], dtype=int)

                    query_pix = np.append(query_pix,
                                          order_pix[query_bool])

                return query_pix
                        
            else:

                return fun(self.nside, nest = self.is_nested,
                           *args, **kwargs)

        return wrapper                

    def query_polygon(self, vertices, inclusive=False, fact=4):
        """
        Returns the pixels whose centers lie within the convex polygon defined 
        by the vertices array (if inclusive is False), or which overlap with 
        this polygon (if inclusive is True).

        Args:
            vertices (float): Vertex array containing the vertices of the 
                polygon, shape (N, 3).
            inclusive (bool): f False, return the exact set of pixels whose 
                pixels centers lie within the region; if True, return all 
                pixels that overlap with the region.
            fact (int): Only used when inclusive=True. The overlapping test 
                will be done at the resolution fact*nside. For NESTED ordering, 
                fact must be a power of 2, less than 2**30, else it can be any 
                positive integer. Default: 4.
            
        Return:
            int array: The pixels which lie within the given polygon.
        """
        
        fun = self._pix_query_fun(hp.query_polygon)

        return fun(vertices, inclusive=inclusive, fact=fact)

    def query_disc(self, vec, radius, inclusive=False, fact=4):
        """

        Args:
            vec (float, sequence of 3 elements): The coordinates of unit vector 
                defining the disk center.
            radius (float): The radius (in radians) of the disk
            inclusive (bool): f False, return the exact set of pixels whose 
                pixels centers lie within the region; if True, return all 
                pixels that overlap with the region.
            fact (int): Only used when inclusive=True. The overlapping test 
                will be done at the resolution fact*nside. For NESTED ordering, 
                fact must be a power of 2, less than 2**30, else it can be any 
                positive integer. Default: 4.
        
        Return:
            int array: The pixels which lie within the given disc.
        """
        
        fun = self._pix_query_fun(hp.query_disc)

        return fun(vec, radius, inclusive=inclusive, fact=fact)

    def query_strip(self,  theta1, theta2, inclusive=False):
        """
        Returns pixels whose centers lie within the colatitude range defined by 
        theta1 and theta2 (if inclusive is False), or which overlap with this 
        region (if inclusive is True). If theta1<theta2, the region between 
        both angles is considered, otherwise the regions 0<theta<theta2 and 
        theta1<theta<pi.

        Args:
            theta (float): First colatitude (radians)
            phi (float): Second colatitude (radians)
            inclusive (bool): f False, return the exact set of pixels whose 
                pixels centers lie within the region; if True, return all 
                pixels that overlap with the region.

        Return:
            int array: The pixels which lie within the given strip.
        """
        
        fun = self._pix_query_fun(hp.query_strip)

        return fun(theta1, theta2, inclusive=inclusive)
    
    def boundaries(self, pix, step = 1):
        """
        Returns an array containing vectors to the boundary of the nominated pixel.

        The returned array has shape (3, 4*step), the elements of which are the 
        x,y,z positions on the unit sphere of the pixel boundary. In order to 
        get vector positions for just the corners, specify step=1.
        """
        
        if self.is_moc:

            def single_pix_bounds(pix):

                nest_pix, order = hmap.uniq2nest(self._uniq[pix])
    
                nside = hp.order2nside(order)
    
                return hp.boundaries(nside, nest_pix, step = step, nest = True)

            moc_bounds = np.vectorize(single_pix_bounds)
            
            return moc_bounds(pix)

        else:

            return hp.boundaries(self.nside, pix, step, nest = self.is_nested)

    
        



# Pixelization related function for multi-order coverage maps
# See http://ivoa.net/documents/MOC

import numpy as np
from numpy import array, log2, sqrt, exp

def uniq2order(uniq):
    """
    Extract the corresponding order from a UNIQ numbered pixel

    Args:
        uniq (int or array): Pixel number

    Return:
        int or array: order
    """

    return (log2(array(uniq)/4)/2).astype(int)

def uniq2nest(uniq):
    """
    Convert from UNIQ ordering scheme to NESTED

    Args:
        uniq (int or array): Pixel number

    Return
        (int or array, int or array): npix,order
    """

    uniq = array(uniq)

    order = uniq2order(uniq)
    npix = uniq - 4 * (4**order)

    return npix,order

def nest2uniq(nside, ipix):
    """
    Convert from from NESTED to UNIQ scheme

    Args:
        nside (int): HEALPix NSIDE parameter
        ipix (int or array): Pixel number in NESTED scheme

    Return:
        int or array
    """

    ipix = array(ipix)

    return 4 * nside * nside + ipix

def nest2range(order_input, pix, order_output):
    """
    Get the equivalent range of pixel that correspond to all
    `child pixels` of a map of a greater order.

    Args:
        order_input (int or array): Order of input pixel
        pix (int or array): Input pixel.
        order_output (int): Order of map with `child pixels`

    Return:
        (int or array, int or array): Start pixel (inclusive) and 
            stop pixel (exclusive) 
    """

    if np.any(order_input > order_output):
        raise ValueError("Input order must be greater or equal to output order")

    npix_ratio = 4 ** (order_output - order_input)

    pix = array(pix)

    return (pix*npix_ratio, (pix+1)*npix_ratio)


def uniq2range(uniq, order):
    """
    Convert from a pixel number in NUNIQ scheme to the range of children 
    pixels that it would correspond to in a NESTED map of a given order

    Args:
       uniq (int or array): Pixel number in NUNIQ scheme
       order (int): Order of equivalent single resolution map

    Return:
        (int or array, int or array): Start pixel (inclusive) and 
            stop pixel (exclusive) 
    """

    pix,pix_order = uniq2nest(uniq)

    return nest2range(pix_order, pix, order)

def range2uniq(order, pix_range):
    """
    Convert from range of children pixels in a NESTED map of a given order
    to the corresponding uniq pixel number.

    Args:
        order (int): Order of equivalent single resolution map
        pix_range (int or array, int or array): Star pixel (inclusive) and 
            stop pixel (exclusive)

    Return:
        int
    """

    npix_ratio = array(pix_range[1]) - array(pix_range[0])

    nside = 2**order / sqrt(npix_ratio)

    pix = array(pix_range[0]) / npix_ratio

    if (nside < 1 or
        any([not float(i).is_integer() for i in array([nside]).flatten()]) or
        any([not float(i).is_integer() for i in array([pix]).flatten()])):
        raise ValueError("pix_range is malformed for a HEALPix "
                         "map of order {}".format(order))

    return nest2uniq(array(nside).astype(int),
                                 array(pix).astype(int))




#!/usr/bin/env python

from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

# Get common version number (https://stackoverflow.com/a/7071358)
import re
VERSIONFILE="HealpixMap/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

setup(name='HealpixMap',
      version = verstr,
      author='Israel Martinez-Castellanos',
      author_email='imc@umd.edu',
      url='https://gitlab.com/burstcube/HealpixMap',
      packages = find_packages(include=["HealpixMap", "HealpixMap.*"]),
      install_requires = ['healpy', 'astropy'],
      description = "Object-oriented healpy wrapper supporting multi-resolution maps",
      long_description = long_description,
      long_description_content_type="text/markdown",
      )

